/// SPDX-License-Identifier: GPL-3.0-or-later
/// SPDX-FileCopyrightText: Copyright © 2016-2025 Tony Garnock-Jones <tonyg@leastfixedpoint.com>
var TokenType;
(function (TokenType) {
    TokenType[TokenType["SPACE"] = 0] = "SPACE";
    TokenType[TokenType["NEWLINE"] = 1] = "NEWLINE";
    TokenType[TokenType["ATOM"] = 2] = "ATOM";
    TokenType[TokenType["STRING"] = 3] = "STRING";
    TokenType[TokenType["OPEN"] = 4] = "OPEN";
    TokenType[TokenType["CLOSE"] = 5] = "CLOSE";
})(TokenType || (TokenType = {}));
function getRange(t) {
    if (Array.isArray(t)) {
        return { start: t[0].start, end: t[t.length - 1].end, synthetic: true };
    }
    else {
        return t;
    }
}
function finishGroup(g, end) {
    return Object.assign(Object.assign({}, g), { end });
}
function makeGroup(open, items, close) {
    return { start: open.start, open, end: close.end, close, items };
}
function isSpace(i) {
    return isTokenType(i, TokenType.SPACE) || isTokenType(i, TokenType.NEWLINE);
}
function isGroup(i) {
    return i && ('items' in i);
}
function isToken(i) {
    return i && ('type' in i);
}
function isTokenType(i, t) {
    return isToken(i) && i.type === t;
}
function foldItems(i, fToken, fGroup, fItems) {
    const walk = (i) => {
        if (isGroup(i)) {
            return fGroup(i, fItems(i.items.map(walk)), walk);
        }
        else {
            return fToken(i);
        }
    };
    return fItems(i.map(walk));
}
function itemText(items, options = {}) {
    return foldItems(items, i => {
        var _a;
        if ((_a = options.color) !== null && _a !== void 0 ? _a : false) {
            switch (i.type) {
                case TokenType.SPACE:
                case TokenType.NEWLINE:
                    return '\x1b[31m' + i.text + '\x1b[0m';
                case TokenType.STRING:
                    return '\x1b[34m' + i.text + '\x1b[0m';
                default:
                    return i.text;
            }
        }
        else {
            return i.text;
        }
    }, (g, inner, k) => { var _a; return k(g.open) + inner + (g.close ? k(g.close) : (_a = options.missing) !== null && _a !== void 0 ? _a : ''); }, strs => strs.join(''));
}

/// SPDX-License-Identifier: GPL-3.0-or-later
/// SPDX-FileCopyrightText: Copyright © 2016-2025 Tony Garnock-Jones <tonyg@leastfixedpoint.com>
function formatPos(p) {
    var _a;
    return p ? `${(_a = p.name) !== null && _a !== void 0 ? _a : '?'}:${p.line}:${p.column}` : '?';
}
function startPos(name) {
    return { line: 1, column: 0, pos: 0, name };
}
function fixPos(p) {
    return Object.assign(Object.assign({}, p), { fixed: true });
}
function advancePos(p, ch) {
    var _a;
    if ((_a = p.fixed) !== null && _a !== void 0 ? _a : false) {
        return ch === '\n';
    }
    else {
        let advancedLine = false;
        p.pos++;
        switch (ch) {
            case '\t':
                p.column = (p.column + 8) & ~7;
                break;
            case '\n':
                p.column = 0;
                p.line++;
                advancedLine = true;
                break;
            case '\r':
                p.column = 0;
                break;
            default:
                p.column++;
                break;
        }
        return advancedLine;
    }
}

/// SPDX-License-Identifier: GPL-3.0-or-later
/// SPDX-FileCopyrightText: Copyright © 2016-2025 Tony Garnock-Jones <tonyg@leastfixedpoint.com>
const alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
const inverse_alphabet = new Map(Object.entries(alphabet).map(([i, c]) => [c, parseInt(i)]));
function vlqDecode(s) {
    var _a;
    let acc = 0;
    let shift_amount = 0;
    const buf = [];
    for (const ch of s) {
        const sextet = (_a = inverse_alphabet.get(ch)) !== null && _a !== void 0 ? _a : 0;
        acc |= (sextet & 0x1f) << shift_amount;
        shift_amount += 5;
        if (!(sextet & 0x20)) {
            const negative = !!(acc & 1);
            acc = acc >> 1;
            if (negative)
                acc = -acc;
            buf.push(acc);
            acc = 0;
            shift_amount = 0;
        }
    }
    return buf;
}
function vlqEncode(ns) {
    const buf = [];
    for (let n of ns) {
        n = (n < 0) ? ((-n) << 1) | 1 : (n << 1);
        do {
            const m = n & 0x1f;
            n = n >> 5;
            const sextet = (n > 0) ? m | 0x20 : m;
            buf.push(alphabet[sextet]);
        } while (n > 0);
    }
    return buf.join('');
}

/// SPDX-License-Identifier: GPL-3.0-or-later
/// SPDX-FileCopyrightText: Copyright © 2016-2025 Tony Garnock-Jones <tonyg@leastfixedpoint.com>
class SpanResult {
    constructor(searchTarget, start) {
        this.items = [];
        this.searchTarget = searchTarget;
        this.start = start;
    }
    get offset() {
        return this.searchTarget - this.start;
    }
    get firstItem() {
        return this.items[0].item;
    }
    get lastItem() {
        return this.items[this.items.length - 1].item;
    }
}
class SpanIndex {
    constructor() {
        this.index = [];
    }
    get(pos) {
        var _a;
        if (this.index.length === 0)
            return null;
        let lo = 0;
        let hi = this.index.length;
        // console.log(`\nsearching for ${pos}`);
        while (true) {
            if (lo === hi) {
                if (lo === 0)
                    return null;
                const e = this.index[lo - 1];
                if (e[0] > pos)
                    throw new Error("INTERNAL ERROR: bad binary search (1)");
                if (((_a = this.index[lo]) === null || _a === void 0 ? void 0 : _a[0]) <= pos)
                    throw new Error("INTERNAL ERROR: bad binary search (2)");
                // console.log(`found ${JSON.stringify(e)}, ${JSON.stringify(this.index[lo] ?? null)}`);
                const r = new SpanResult(pos, e[0]);
                e[1].forEach(([end, item]) => {
                    if (pos <= end) {
                        r.items.push({ end, item });
                    }
                });
                return (r.items.length > 0) ? r : null;
            }
            const mid = (lo + hi) >> 1;
            const e = this.index[mid];
            // console.log(`${pos} lo ${lo} hi ${hi} mid ${mid} probe ${JSON.stringify(e)}`);
            if (e[0] <= pos) {
                lo = mid + 1;
            }
            else {
                hi = mid;
            }
        }
    }
}
class SpanInfo {
    constructor() {
        this.spans = new Map();
    }
    add(start, end, t) {
        if (!this.spans.has(start)) {
            this.spans.set(start, []);
        }
        this.spans.get(start).push([end, t]);
    }
    index() {
        const i = new SpanIndex();
        this.spans.forEach((ends, start) => {
            ends.sort((a, b) => a[0] - b[0]);
            i.index.push([start, ends]);
        });
        i.index.sort((a, b) => a[0] - b[0]);
        return i;
    }
}

/// SPDX-License-Identifier: GPL-3.0-or-later
function encodeMapping(entry) {
    const a = [entry.generatedStartColumn];
    if ('sourceIndex' in entry) {
        a.push(entry.sourceIndex);
        a.push(entry.sourceStartLine);
        a.push(entry.sourceStartColumn);
        if ('nameIndex' in entry) {
            a.push(entry.nameIndex);
        }
    }
    return a;
}
function maybeDelta(newValue, oldValue) {
    // console.log('maybeDelta', oldValue, newValue);
    return (oldValue === void 0) ? newValue : newValue - oldValue;
}
class CodeWriter {
    constructor(file) {
        var _a;
        this.sources = [];
        this.chunks = [];
        this.mappings = [];
        this.targetToSourceMap = new SpanInfo();
        this.sourceToTargetMap = new SpanInfo();
        this.previous = {};
        this.previousPos = null;
        this.file = file;
        this.pos = startPos((_a = this.file) !== null && _a !== void 0 ? _a : '');
    }
    get text() {
        return this.chunks.join('');
    }
    get map() {
        // console.log(this.mappings.map(segs => segs.map(encodeMapping)));
        const mappings = this.mappings.map(segments => segments.map(encodeMapping).map(vlqEncode).join(',')).join(';');
        const m = {
            version: 3,
            sources: [...this.sources],
            names: [],
            mappings,
        };
        if (this.file !== null)
            m.file = this.file;
        return m;
    }
    finishLine() {
        // console.log('newline');
        this.mappings.push([]);
        this.previous.generatedStartColumn = undefined;
        this.previousPos = null;
    }
    sourceIndexFor(name) {
        let i = this.sources.indexOf(name);
        if (i === -1) {
            this.sources.push(name);
            i = this.sources.length - 1;
        }
        return i;
    }
    addMapping(p, type) {
        // console.log('considering', p, type);
        const oldPos = this.previousPos;
        if ((oldPos === null || oldPos.name === p.name) &&
            (type === TokenType.SPACE || type === TokenType.NEWLINE)) {
            // console.log('whitespace skip');
            if (this.previousPos !== null) {
                this.previousPos = p;
            }
            return;
        }
        this.previousPos = p;
        if (((oldPos === null || oldPos === void 0 ? void 0 : oldPos.name) === p.name) &&
            ((p.name === null) ||
                (((oldPos === null || oldPos === void 0 ? void 0 : oldPos.column) === p.column) && ((oldPos === null || oldPos === void 0 ? void 0 : oldPos.line) === p.line)))) {
            // console.log('skipping', this.previous, oldPos, p);
            return;
        }
        let n = {
            generatedStartColumn: maybeDelta(this.pos.column, this.previous.generatedStartColumn),
        };
        this.previous.generatedStartColumn = this.pos.column;
        if (p.name !== null) {
            const sourceIndex = this.sourceIndexFor(p.name);
            n = Object.assign(Object.assign({}, n), { sourceIndex: maybeDelta(sourceIndex, this.previous.sourceIndex), sourceStartColumn: maybeDelta(p.column, this.previous.sourceStartColumn), sourceStartLine: maybeDelta(p.line - 1, this.previous.sourceStartLine) });
            this.previous.sourceIndex = sourceIndex;
            this.previous.sourceStartColumn = p.column;
            this.previous.sourceStartLine = p.line - 1;
        }
        // console.log('pushing',
        //             n,
        //             this.previous,
        //             oldPos?.line + ':' + oldPos?.column,
        //             p.line + ':' + p.column);
        this.mappings[this.mappings.length - 1].push(n);
    }
    chunk(p, s, type) {
        p = Object.assign({}, p);
        this.chunks.push(s);
        if (this.mappings.length === 0)
            this.finishLine();
        this.addMapping(p, type);
        for (const ch of s) {
            advancePos(p, ch);
            if (advancePos(this.pos, ch)) {
                this.finishLine();
                this.addMapping(p, type);
            }
        }
    }
    emit(i) {
        if (Array.isArray(i)) {
            i.forEach(j => this.emit(j));
        }
        else if (isGroup(i)) {
            this.emit(i.open);
            this.emit(i.items);
            if (i.close)
                this.emit(i.close);
        }
        else if (i === null) ;
        else {
            const targetStart = this.pos.pos;
            if (!i.synthetic)
                this.sourceToTargetMap.add(i.start.pos, i.end.pos, targetStart);
            this.chunk(i.start, i.text, i.type);
            this.targetToSourceMap.add(targetStart, this.pos.pos, i);
        }
    }
}

/// SPDX-License-Identifier: GPL-3.0-or-later
/// SPDX-FileCopyrightText: Copyright © 2016-2025 Tony Garnock-Jones <tonyg@leastfixedpoint.com>
function atEnd(xs) {
    return xs.item === null;
}
function notAtEnd(xs) {
    return xs.item !== null;
}
class ArrayList {
    constructor(items, context, index = 0) {
        this.context = context;
        this.index = 0;
        this.items = items;
        this.index = index;
    }
    get item() {
        var _a;
        return (_a = this.items[this.index]) !== null && _a !== void 0 ? _a : null;
    }
    get next() {
        if (this.index >= this.items.length)
            return null;
        return new ArrayList(this.items, this.context, this.index + 1);
    }
    toArray() {
        return this.items.slice(this.index);
    }
    [Symbol.iterator]() {
        let i = this;
        return {
            next() {
                if (notAtEnd(i)) {
                    const value = i.item;
                    i = i.next;
                    return { done: false, value };
                }
                else {
                    return { done: true, value: null };
                }
            }
        };
    }
}

/// SPDX-License-Identifier: GPL-3.0-or-later
function match(p, items, failure, context) {
    const r = p(new ArrayList(items, context));
    if (r === null)
        return failure;
    if (notAtEnd(skipSpace(r[1])))
        return failure;
    return r[0];
}
const noItems = (c) => new ArrayList([], c);
const fail = _i => null;
function succeed(t) { return i => [t, i]; }
const discard = i => [void 0, noItems(i.context)];
const rest = i => [i.toArray(), noItems(i.context)];
const end = i => atEnd(skipSpace(i)) ? [void 0, noItems(i.context)] : null;
const pos = i => notAtEnd(i) ? [i.item.start, i] : null;
const newline = i => {
    while (notAtEnd(i) && isTokenType(i.item, TokenType.SPACE))
        i = i.next;
    if (!notAtEnd(i) || !isTokenType(i.item, TokenType.NEWLINE))
        return null;
    return [i.item, i.next];
};
function skipSpace(i) {
    while (notAtEnd(i) && isSpace(i.item))
        i = i.next;
    return i;
}
function collectSpace(i, acc) {
    while (notAtEnd(i) && isSpace(i.item)) {
        acc.push(i.item);
        i = i.next;
    }
    return i;
}
function withoutSpace(p) {
    return i => p(skipSpace(i));
}
function not(p, v) {
    return i => p(i) === null ? [v, i] : null;
}
function follows(p) {
    return i => {
        const r = p(i);
        if (r === null)
            return null;
        return [r[0], i];
    };
}
function seq(...patterns) {
    return i => {
        for (const p of patterns) {
            const r = p(i);
            if (r === null)
                return null;
            i = r[1];
        }
        return [void 0, i];
    };
}
function seqTuple(...patterns) {
    return i => {
        const rs = [];
        for (const p of patterns) {
            const r = p(i);
            if (r === null)
                return null;
            rs.push(r[0]);
            i = r[1];
        }
        return [rs, i];
    };
}
function alt(...alts) {
    return i => {
        for (const a of alts) {
            const r = a(i);
            if (r !== null)
                return r;
        }
        return null;
    };
}
function scope(pf) {
    return i => {
        const scope = Object.create(null);
        const r = pf(scope)(i);
        if (r === null)
            return null;
        return [scope, r[1]];
    };
}
function value(pf) {
    return i => {
        const scope = Object.create(null);
        const r = pf(scope)(i);
        if (r === null)
            return null;
        return [scope.value, r[1]];
    };
}
function bind(target, key, pattern) {
    return i => {
        const r = pattern(i);
        if (r === null)
            return null;
        target[key] = r[0];
        return r;
    };
}
function exec(thunk) {
    return i => {
        thunk(i);
        return [void 0, i];
    };
}
function map(p, f) {
    return i => {
        const r = p(i);
        if (r === null)
            return null;
        return [f(r[0]), r[1]];
    };
}
function mapm(p, f) {
    return i => {
        const r = p(i);
        if (r === null)
            return null;
        return f(r[0])(r[1]);
    };
}
function group(opener, items, options = {}) {
    return i => {
        var _a, _b;
        if ((_a = options.skipSpace) !== null && _a !== void 0 ? _a : true)
            i = skipSpace(i);
        if (!notAtEnd(i))
            return null;
        if (!isGroup(i.item))
            return null;
        if (i.item.open.text !== opener)
            return null;
        const r = items(new ArrayList(i.item.items, opener));
        if (r === null)
            return null;
        if (!atEnd(r[1]))
            return null;
        return [r[0], ((_b = options.advance) !== null && _b !== void 0 ? _b : true) ? i.next : i];
    };
}
function atomString(text, options = {}) {
    return map(atom(text, options), _t => text);
}
function atom(text, options = {}) {
    return i => {
        var _a, _b, _c;
        if ((_a = options.skipSpace) !== null && _a !== void 0 ? _a : true)
            i = skipSpace(i);
        if (!notAtEnd(i))
            return null;
        if (!isToken(i.item))
            return null;
        if (i.item.type !== ((_b = options.tokenType) !== null && _b !== void 0 ? _b : TokenType.ATOM))
            return null;
        if (text !== void 0 && i.item.text !== text)
            return null;
        return [i.item, ((_c = options.advance) !== null && _c !== void 0 ? _c : true) ? i.next : i];
    };
}
function anything(options = {}) {
    return i => {
        var _a, _b;
        if ((_a = options.skipSpace) !== null && _a !== void 0 ? _a : true)
            i = skipSpace(i);
        if (!notAtEnd(i))
            return null;
        return [i.item, ((_b = options.advance) !== null && _b !== void 0 ? _b : true) ? i.next : i];
    };
}
function upTo(p) {
    return i => {
        const acc = [];
        while (true) {
            const r = p(i);
            if (r !== null)
                return [acc, i];
            if (!notAtEnd(i))
                break;
            acc.push(i.item);
            i = i.next;
        }
        return null;
    };
}
function separatedBy(itemPattern, separator) {
    return i => {
        const acc = [];
        if (end(i) !== null)
            return [acc, noItems(i.context)];
        while (true) {
            {
                const r = itemPattern(i);
                if (r === null)
                    return null;
                acc.push(r[0]);
                i = r[1];
            }
            {
                const r = separator(i);
                if (r === null) {
                    if (end(i) !== null)
                        return [acc, noItems(i.context)];
                    return null;
                }
                i = r[1];
            }
        }
    };
}
function separatedOrTerminatedBy(itemPattern, separator) {
    return i => {
        const acc = [];
        if (end(i) !== null)
            return [acc, noItems(i.context)];
        while (true) {
            {
                const r = itemPattern(i);
                if (r === null)
                    return null;
                acc.push(r[0]);
                i = r[1];
            }
            {
                const r = separator(i);
                if (r === null) {
                    if (end(i) !== null)
                        return [acc, noItems(i.context)];
                    return null;
                }
                else {
                    i = r[1];
                    if (end(i) !== null)
                        return [acc, noItems(i.context)];
                }
            }
        }
    };
}
function repeat(p, options = {}) {
    return i => {
        var _a;
        const acc = [];
        let needSeparator = false;
        const finish = () => { var _a; return (acc.length < ((_a = options.min) !== null && _a !== void 0 ? _a : 0)) ? null : [acc, i]; };
        while (true) {
            if (acc.length == ((_a = options.max) !== null && _a !== void 0 ? _a : Infinity))
                return [acc, i];
            if (needSeparator) {
                if (options.separator) {
                    const r = options.separator(i);
                    if (r === null)
                        return finish();
                    i = r[1];
                }
            }
            else {
                needSeparator = true;
            }
            const r = p(i);
            if (r === null)
                return finish();
            acc.push(r[0]);
            i = r[1];
        }
    };
}
function option(p) {
    return repeat(p, { max: 1 });
}
//---------------------------------------------------------------------------
// Search-and-replace over Item
function replace(items, outerContext, p, f, end = items.length > 0 ? items[items.length - 1].end : startPos(null)) {
    const walkItems = (items, end, context) => {
        let i = new ArrayList(items, context);
        const acc = [];
        while (notAtEnd(i = collectSpace(i, acc))) {
            const r = p(i);
            if (r !== null) {
                acc.push(...f(r[0], notAtEnd(i) ? i.item.start : end, notAtEnd(r[1]) ? r[1].item.start : end));
                i = r[1];
            }
            else if (isToken(i.item)) {
                acc.push(i.item);
                i = i.next;
            }
            else {
                acc.push(Object.assign(Object.assign({}, i.item), { items: walkItems(i.item.items, i.item.end, i.item.open.text) }));
                i = i.next;
            }
        }
        return acc;
    };
    return walkItems(items, end, outerContext);
}

/// SPDX-License-Identifier: GPL-3.0-or-later
class Scanner {
    constructor(pos, synthetic) {
        this.charBuffer = null;
        this.tokenBuffer = null;
        this.delimiters = ' \t\n\r\'"`.,;()[]{}/\\';
        this.pos = Object.assign({}, pos);
        this.synthetic = synthetic;
    }
    [Symbol.iterator]() {
        return this;
    }
    peekChar() {
        if (this.charBuffer !== null)
            return this.charBuffer;
        this.charBuffer = this._peekChar();
        return this.charBuffer;
    }
    dropChar() {
        if (this.charBuffer === null)
            this.peekChar();
        if (this.charBuffer !== null) {
            this._dropChar();
            advancePos(this.pos, this.charBuffer);
            this.charBuffer = null;
        }
    }
    shiftChar() {
        const ch = this.peekChar();
        this.dropChar();
        return ch;
    }
    makeToken(start, type, text) {
        return Object.assign({ type, start, end: this.mark(), text }, this.synthetic && { synthetic: true });
    }
    makeGroupInProgress(open, items = []) {
        return Object.assign({ start: open.start, open, close: null, items }, this.synthetic && { synthetic: true });
    }
    mark() {
        return Object.assign({}, this.pos);
    }
    _while(pred, f) {
        while (true) {
            const ch = this.peekChar();
            if (!pred(ch))
                return;
            this.dropChar();
            f(ch);
        }
    }
    _collectSpace(buf = '', start = this.mark()) {
        this._while(ch => ch !== null && this.isSpace(ch), ch => buf = buf + ch);
        return this.makeToken(start, TokenType.SPACE, buf);
    }
    _punct(type) {
        return this.makeToken(this.mark(), type, this.shiftChar());
    }
    _str(q, buf, isTemplate) {
        const start = this.mark();
        let ch;
        while (true) {
            ch = this.shiftChar();
            if (ch !== null)
                buf = buf + ch;
            if (ch === null ||
                ch === q ||
                (isTemplate ? (ch === '$' && this.peekChar() === '{') : (ch === '\n'))) {
                return this.makeToken(start, TokenType.STRING, buf);
            }
            if (ch === '\\') {
                ch = this.shiftChar();
                if (ch === '\n') ;
                else if (ch !== null) {
                    buf = buf + ch;
                }
            }
        }
    }
    templateConstantFragment() {
        if (this.tokenBuffer !== null)
            throw new Error("Internal error: templateConstantFragment");
        const t = this._str('`', '', true);
        return (t.text.length === 0) ? null : t;
    }
    isSpace(ch) {
        return ' \t\r'.indexOf(ch) >= 0;
    }
    isDelimiter(ch) {
        return this.delimiters.indexOf(ch) >= 0;
    }
    addDelimiters(newDelimiters) {
        this.delimiters = this.delimiters + newDelimiters;
    }
    _atom(start = this.mark(), buf = '') {
        let ch;
        while (true) {
            ch = this.peekChar();
            if (ch === null || this.isDelimiter(ch)) {
                return this.makeToken(start, TokenType.ATOM, buf);
            }
            buf = buf + ch;
            this.dropChar();
        }
    }
    _maybeComment() {
        const start = this.mark();
        let buf = this.shiftChar();
        let ch = this.peekChar();
        if (ch === null)
            return this._collectSpace(buf, start);
        switch (ch) {
            case '/': // single-line comment.
                this._while(ch => ch !== null && ch !== '\n', ch => buf = buf + ch);
                return this._collectSpace(buf, start);
            case '*': // delimited comment.
                {
                    let seenStar = false;
                    buf = buf + this.shiftChar();
                    while (true) {
                        ch = this.shiftChar();
                        if (ch === null)
                            break;
                        buf = buf + ch;
                        if ((ch === '/') && seenStar)
                            break;
                        seenStar = (ch === '*');
                    }
                    return this._collectSpace(buf, start);
                }
            default:
                return this._atom(start, buf);
        }
    }
    _peek() {
        let ch = this.peekChar();
        if (ch === null)
            return null;
        switch (ch) {
            case ' ':
            case '\t':
            case '\r':
                return this._collectSpace();
            case '\n':
                return this._punct(TokenType.NEWLINE);
            case '(':
            case '[':
            case '{':
                return this._punct(TokenType.OPEN);
            case ')':
            case ']':
            case '}':
                return this._punct(TokenType.CLOSE);
            case '\'':
            case '"':
                this.dropChar();
                return this._str(ch, ch, false);
            case '`':
                this.dropChar();
                return this._str(ch, ch, true);
            case '/':
                return this._maybeComment();
            default:
                if (this.isDelimiter(ch)) {
                    return this._punct(TokenType.ATOM);
                }
                else {
                    return this._atom(this.mark(), this.shiftChar());
                }
        }
    }
    peek() {
        if (this.tokenBuffer === null)
            this.tokenBuffer = this._peek();
        return this.tokenBuffer;
    }
    drop() {
        if (this.tokenBuffer === null)
            this.peek();
        this.tokenBuffer = null;
    }
    shift() {
        const t = this.peek();
        this.drop();
        return t;
    }
    next() {
        const t = this.shift();
        if (t === null) {
            return { done: true, value: null };
        }
        else {
            return { done: false, value: t };
        }
    }
}
class StringScanner extends Scanner {
    constructor(pos, input, synthetic) {
        super(pos, synthetic);
        this.input = input;
        this.index = 0;
    }
    _peekChar() {
        var _a;
        return (_a = this.input[this.index]) !== null && _a !== void 0 ? _a : null;
    }
    _dropChar() {
        this.index++;
    }
}

/// SPDX-License-Identifier: GPL-3.0-or-later
function matchingParen(c) {
    switch (c) {
        case ')': return '(';
        case ']': return '[';
        case '}': return '{';
        default: return null;
    }
}
class LaxReader {
    constructor(scanner) {
        this.stack = [];
        this.scanner = scanner;
    }
    [Symbol.iterator]() {
        return this;
    }
    stackTop() {
        var _a;
        return (_a = this.stack[this.stack.length - 1]) !== null && _a !== void 0 ? _a : null;
    }
    popUntilMatch(t) {
        const m = matchingParen(t.text);
        if (m !== null && !this.stack.some(g => g.open.text === m)) {
            if (this.stack.length > 0) {
                this.stackTop().items.push(t);
                return 'continue';
            }
        }
        else {
            while (this.stack.length > 0) {
                const inner = finishGroup(this.stack.pop(), t.end);
                if (inner.open.text === m) {
                    inner.close = t;
                }
                if (this.stack.length === 0) {
                    return inner;
                }
                else {
                    const outer = this.stackTop();
                    outer.items.push(inner);
                    if (inner.open.text === m) {
                        return 'continue';
                    }
                }
            }
        }
        return 'eof';
    }
    _eofClose() {
        return this.scanner.makeToken(this.scanner.mark(), TokenType.CLOSE, '');
    }
    peek() {
        var _a;
        return (_a = this.scanner.peek()) !== null && _a !== void 0 ? _a : this._eofClose();
    }
    drop() {
        this.scanner.drop();
    }
    inTemplateString() {
        const i = this.stackTop();
        return (i !== null) && (i.open.text[0] === '`');
    }
    expectTemplateVariablePart() {
        const variablePartOpen = this.peek();
        if (variablePartOpen.type !== TokenType.OPEN || variablePartOpen.text !== '{') {
            throw new Error("Internal parser error: template string variable part mismatch");
        }
        this.drop();
        this.stack.push(this.scanner.makeGroupInProgress(variablePartOpen));
    }
    read() {
        var _a;
        while (true) {
            if (this.inTemplateString()) {
                const t = (_a = this.scanner.templateConstantFragment()) !== null && _a !== void 0 ? _a : this._eofClose();
                switch (t.type) {
                    case TokenType.CLOSE: // always means eof, in this case
                        return this.popUntilMatch(t); // will always be a Group
                    case TokenType.STRING:
                        if (t.text[t.text.length - 1] === '`') {
                            const inner = finishGroup(this.stack.pop(), t.end);
                            inner.close = t;
                            if (this.stack.length === 0) {
                                return inner;
                            }
                            else {
                                this.stackTop().items.push(inner);
                                break;
                            }
                        }
                        else {
                            // Another unclosed fragment. Expect {...} again.
                            this.stackTop().items.push(t);
                            this.expectTemplateVariablePart();
                            break;
                        }
                    default:
                        throw new Error("Internal error: LaxReader.read()");
                }
            }
            else {
                let g = this.stackTop();
                const t = this.peek();
                switch (t.type) {
                    case TokenType.STRING:
                        if (t.text[0] === '`' && t.text[t.text.length - 1] !== '`') {
                            // Unclosed template string - so expect {...} followed by more
                            // fragments. Encode this as a group with opener the first constant
                            // template fragment and closer the last template fragment.
                            this.drop();
                            this.stack.push(this.scanner.makeGroupInProgress(t));
                            // We know, from the implementation of the scanner, that the current
                            // character is '{'.
                            this.expectTemplateVariablePart();
                            break;
                        }
                    case TokenType.SPACE:
                    case TokenType.NEWLINE:
                    case TokenType.ATOM:
                        this.drop();
                        if (g === null) {
                            return t;
                        }
                        g.items.push(t);
                        break;
                    case TokenType.OPEN:
                        this.drop();
                        this.stack.push(this.scanner.makeGroupInProgress(t));
                        break;
                    case TokenType.CLOSE: {
                        this.drop();
                        const i = this.popUntilMatch(t);
                        if (i === 'eof')
                            return null;
                        if (i === 'continue')
                            break;
                        return i;
                    }
                }
            }
        }
    }
    readToEnd() {
        return Array.from(this);
    }
    next() {
        const i = this.read();
        if (i === null) {
            return { done: true, value: null };
        }
        else {
            return { done: false, value: i };
        }
    }
}
function laxRead(source, options = {}) {
    var _a, _b;
    const start = (_a = options.start) !== null && _a !== void 0 ? _a : startPos((_b = options.name) !== null && _b !== void 0 ? _b : null);
    const scanner = new StringScanner(start, source, options.synthetic);
    if (options.extraDelimiters)
        scanner.addDelimiters(options.extraDelimiters);
    const reader = new LaxReader(scanner);
    return reader.readToEnd();
}

/// SPDX-License-Identifier: GPL-3.0-or-later
const substPat = scope((o) => seq(atom('$'), seq(bind(o, 'pos', pos), group('{', end, { skipSpace: false }))));
function toItems(readOptions, s, pos) {
    return typeof s === 'string' ? laxRead(s, Object.assign(Object.assign({}, readOptions), { start: pos, synthetic: true })) : s;
}
class Templates {
    constructor(defaultPos = startPos(null), readOptions = {}) {
        this.sources = {};
        this.recordSources = false;
        this.defaultPos = defaultPos;
        this.readOptions = readOptions;
    }
    template(start0 = this.defaultPos, context = null) {
        const start = (typeof start0 === 'string') ? startPos(start0) : start0;
        return (consts, ...vars) => {
            var _a;
            const sourcePieces = [consts[0]];
            for (let i = 1; i < consts.length; i++) {
                sourcePieces.push('${}');
                sourcePieces.push(consts[i]);
            }
            const source = sourcePieces.join('');
            if (this.recordSources) {
                if (start.name !== null) {
                    if (start.name in this.sources && this.sources[start.name] !== source) {
                        throw new Error(`Duplicate template name: ${start.name}`);
                    }
                    this.sources[start.name] = source;
                }
            }
            let i = 0;
            return replace(laxRead(source, Object.assign(Object.assign({}, this.readOptions), { start, extraDelimiters: ((_a = this.readOptions.extraDelimiters) !== null && _a !== void 0 ? _a : '') + '$', synthetic: true })), context, substPat, sub => toItems(this.readOptions, vars[i++], sub.pos));
        };
    }
    sourceFor(name) {
        return this.sources[name];
    }
}
function joinItems(itemss, separator0 = '', readOptions = {}) {
    if (itemss.length === 0)
        return [];
    const separator = toItems(readOptions, separator0, startPos(null));
    const acc = [...itemss[0]];
    for (let i = 1; i < itemss.length; i++) {
        acc.push(...separator, ...itemss[i]);
    }
    return acc;
}
function commaJoin(itemss) {
    return joinItems(itemss, ', ');
}
const anonymousTemplate = (new Templates()).template();
// const lib = new Templates();
// const t = (o: {xs: Items}) => lib.template('testTemplate')`YOYOYOYO ${o.xs}><`;
// console.log(t({xs: lib.template()`hello there`}));

/// SPDX-License-Identifier: GPL-3.0-or-later

var index = /*#__PURE__*/Object.freeze({
    __proto__: null,
    CodeWriter: CodeWriter,
    atEnd: atEnd,
    notAtEnd: notAtEnd,
    ArrayList: ArrayList,
    match: match,
    noItems: noItems,
    fail: fail,
    succeed: succeed,
    discard: discard,
    rest: rest,
    end: end,
    pos: pos,
    newline: newline,
    skipSpace: skipSpace,
    collectSpace: collectSpace,
    withoutSpace: withoutSpace,
    not: not,
    follows: follows,
    seq: seq,
    seqTuple: seqTuple,
    alt: alt,
    scope: scope,
    value: value,
    bind: bind,
    exec: exec,
    map: map,
    mapm: mapm,
    group: group,
    atomString: atomString,
    atom: atom,
    anything: anything,
    upTo: upTo,
    separatedBy: separatedBy,
    separatedOrTerminatedBy: separatedOrTerminatedBy,
    repeat: repeat,
    option: option,
    replace: replace,
    formatPos: formatPos,
    startPos: startPos,
    fixPos: fixPos,
    advancePos: advancePos,
    LaxReader: LaxReader,
    laxRead: laxRead,
    Scanner: Scanner,
    StringScanner: StringScanner,
    SpanResult: SpanResult,
    SpanIndex: SpanIndex,
    SpanInfo: SpanInfo,
    Templates: Templates,
    joinItems: joinItems,
    commaJoin: commaJoin,
    anonymousTemplate: anonymousTemplate,
    get TokenType () { return TokenType; },
    getRange: getRange,
    finishGroup: finishGroup,
    makeGroup: makeGroup,
    isSpace: isSpace,
    isGroup: isGroup,
    isToken: isToken,
    isTokenType: isTokenType,
    foldItems: foldItems,
    itemText: itemText,
    vlqDecode: vlqDecode,
    vlqEncode: vlqEncode
});

/// SPDX-License-Identifier: GPL-3.0-or-later
//---------------------------------------------------------------------------
// Parsers
function kw(text) {
    return value(o => seq(atom(':'), bind(o, 'value', atom(text, { skipSpace: false }))));
}
const DOUBLE_QUOTE = anonymousTemplate `"`;
class SyndicateParser {
    constructor() {
        this.statementBoundary = alt(atom(';'), newline, seq(end, i => {
            if (i.context === null || i.context === '{')
                return discard(i);
            // ^ toplevel, or inside braces, so presumably statement context
            return fail(); // otherwise, parens or brackets presumably, so not statement context
        }));
        this.exprBoundary = alt(atom(';'), atom(','), group('{', discard), end);
        this.identifier = atom();
        this.defaultBinder = this.binder();
        this.type = this.expr;
        this.headerExpr = this.expr(kw('asserting'), kw('let'));
        // Principal: Turn
        this.spawn = this.turnAction(o => {
            o.linkedToken = null;
            o.parentBinders = [];
            o.parentInits = [];
            o.body = [];
            return seq(atom('spawn'), option(map(atom('linked'), tok => o.linkedToken = tok)), option(seq(atom('named'), bind(o, 'name', this.headerExpr))), repeat(alt(
            /* seq(kw('asserting'), map(this.headerExpr, e => o.initialAssertions.push(e))), */
            map(scope((l) => seq(kw('let'), bind(l, 'b', this.defaultBinder), atom('='), bind(l, 'init', this.headerExpr))), l => {
                o.parentBinders.push(l.b);
                o.parentInits.push(l.init);
            }))), this.block(o.body));
        });
        // Principal: Turn
        this.fieldDeclarationStatement = this.turnAction(o => {
            return seq(atom('field'), bind(o, 'field', this.defaultBinder), option(seq(atom('='), bind(o, 'init', this.expr()))), this.statementBoundary);
        });
        // Principal: Turn
        this.assertionEndpointStatement = this.turnAction(o => {
            o.isDynamic = true;
            return seq(atom('assert'), option(map(kw('snapshot'), _ => o.isDynamic = false)), bind(o, 'template', this.expr(seq(atom('when'), group('(', discard)))), option(seq(atom('when'), group('(', bind(o, 'test', this.expr())))), this.statementBoundary);
        });
        // Principal: Turn
        this.dataflowStatement = this.blockTurnAction(atom('dataflow'));
        // Principal: Turn
        this.eventHandlerEndpointStatement = this.turnAction(o => {
            o.facetToStop = 'none';
            o.once = false;
            o.isDynamic = true;
            o.body = [];
            return seq(alt(seq(option(seq(atom('stop'), map(option(this.expr1(atom('on'))), es => {
                o.facetToStop = es.length ? es[0] : 'default';
            }))), atom('on')), map(atom('once'), _ => {
                o.once = true;
                o.facetToStop = 'once-wrapper';
            })), alt(seq(map(group('(', bind(o, 'predicate', this.expr())), _ => o.triggerType = 'dataflow'), this.mandatoryIfNotTerminal(o, this.statement(o.body))), mapm(seq(bind(o, 'triggerType', atomString('stop')), option(this.statement(o.body))), v => ((o.facetToStop !== 'none') || o.once) ? fail : succeed(v)), seq(bind(o, 'triggerType', alt(atomString('asserted'), atomString('retracted'), atomString('message'))), option(map(kw('snapshot'), _ => o.isDynamic = false)), bind(o, 'pattern', this.valuePattern(1, atom('=>'), seq(atom('when'), group('(', discard)))), option(seq(atom('when'), group('(', bind(o, 'test', this.expr())))), this.mandatoryIfNotTerminal(o, seq(atom('=>'), this.statement(o.body))))));
        });
        // Principal: none
        this.typeDefinitionStatement = scope(o => seq(bind(o, 'expectedUse', alt(atomString('message'), atomString('assertion'))), atom('type'), bind(o, 'label', this.identifier), group('(', bind(o, 'fields', repeat(this.defaultBinder, { separator: atom(',') }))), option(seq(atom('='), bind(o, 'wireName', withoutSpace(upTo(this.statementBoundary))))), this.statementBoundary));
        // Principal: Turn
        this.messageSendStatement = this.turnAction(o => seq(atom('send'), atom('message'), not(this.statementBoundary), bind(o, 'expr', withoutSpace(upTo(this.statementBoundary))), this.statementBoundary));
        // Principal: Turn
        this.duringStatement = this.turnAction(o => {
            o.body = [];
            return seq(atom('during'), bind(o, 'pattern', this.valuePattern(1, atom('=>'), seq(atom('when'), group('(', discard)))), option(seq(atom('when'), group('(', bind(o, 'test', this.expr())))), seq(atom('=>'), this.statement(o.body)));
        });
        // Principal: Turn
        this.reactStatement = this.turnAction(o => {
            o.label = null;
            o.body = [];
            return seq(option(map(seqTuple(this.identifier, atom(':')), ([i, _colon]) => o.label = i)), atom('react'), this.block(o.body));
        });
        // Principal: Turn
        this.stopStatement = this.turnAction(o => {
            o.facetToStop = 'default';
            o.body = [];
            return seq(atom('stop'), option(map(this.expr1(), e => o.facetToStop = e)), alt(this.block(o.body), this.statementBoundary));
        });
        // Principal: Turn
        this.syncStatement = this.turnAction(o => {
            o.body = [];
            return seq(atom('sync'), map(this.expr1(), e => o.peerToSyncWith = e), this.block(o.body));
        });
        // Principal: none
        this.atStatement = scope(o => {
            o.body = [];
            return seq(atom('at'), bind(o, 'target', this.expr()), this.block(o.body));
        });
        // Principal: none
        this.createExpression = scope(o => seq(atom('create'), bind(o, 'entity', this.expr())));
        //---------------------------------------------------------------------------
        // Syntax of patterns over Value, used in endpoints
        // $id - capture of discard
        // _ - discard
        //
        // expr(pat, ...) - record ctor
        // $id(pat) - nested capture
        // [pat, ...] - array pat
        // {expr: pat, ...} - dict pat
        //
        // expr(expr, ...) - constant
        // [expr, ...] - constant
        // {expr: expr, ...} - constant
        // other - constant
        this.pCaptureBinder = (b) => mapm(b, i => {
            if (i.id.text.startsWith('$')) {
                const adjustedStart = Object.assign({}, i.id.start);
                advancePos(adjustedStart, ' ');
                const adjustedId = Object.assign(Object.assign({}, i.id), { start: adjustedStart, text: i.id.text.slice(1) });
                return succeed({ id: adjustedId, type: i.type });
            }
            else {
                return fail;
            }
        });
        this.pCaptureDefaultBinder = this.pCaptureBinder(this.defaultBinder);
        this.pDiscard = mapm(this.identifier, i => i.text === '_' ? succeed(void 0) : fail);
    }
    block(acc) {
        return group('{', map(rest, items => (acc === null || acc === void 0 ? void 0 : acc.push(...items), items)));
    }
    binder(..._extraStops) { return scope(o => bind(o, 'id', this.identifier)); }
    expr(...extraStops) {
        return withoutSpace(upTo(alt(this.exprBoundary, ...extraStops)));
    }
    expr1(...extraStops) {
        return mapm(this.expr(...extraStops), e => e.length ? succeed(e) : fail);
    }
    propertyNameExpr() {
        return alt(map(atom(), name => [...DOUBLE_QUOTE, name, ...DOUBLE_QUOTE]), map(atom(void 0, { tokenType: TokenType.STRING }), str => [str]), group('[', this.expr()));
    }
    statement(acc) {
        return alt(this.block(acc), withoutSpace(seq(map(upTo(this.statementBoundary), items => acc.push(...items)), map(this.statementBoundary, i => i ? acc.push(i) : void 0))));
    }
    turnAction(pattern) {
        return i => {
            const scope = Object.create(null);
            const p = pattern(scope);
            const r = p(i);
            if (r === null)
                return null;
            return [scope, r[1]];
        };
    }
    blockTurnAction(kw) {
        return this.turnAction(o => {
            o.body = [];
            return seq(kw, this.block(o.body));
        });
    }
    // Expects o to parse something statement-like.
    mandatoryIfNotTerminal(o, p) {
        return i => {
            return (o.facetToStop !== 'none') ? alt(p, this.statementBoundary)(i) : p(i);
        };
    }
    pUnquote(level, extraStops) {
        if (level == 1) {
            return scope(o => {
                o.type = 'POuterUnquote';
                return seq(atom('\\'), bind(o, 'outer', this.expr(...extraStops)));
            });
        }
        else {
            return scope(o => {
                o.type = 'PUnquote';
                return seq(atom('\\'), bind(o, 'unquoted', this.valuePattern(level - 1, ...extraStops)));
            });
        }
    }
    pQuote(level, extraStops) {
        return scope(o => {
            o.type = 'PQuote';
            return seq(kw('pattern'), bind(o, 'quoted', this.valuePattern(level + 1, ...extraStops)));
        });
    }
    hasCapturesOrDiscards(e) {
        return foldItems(e, t => match(alt(this.pCaptureDefaultBinder, this.pDiscard), [t], null, '(') !== null, (_g, b, _k) => b, bs => bs.some(b => b));
    }
    pArray(level) {
        return scope(o => {
            o.type = 'PArray';
            return group('[', mapm(bind(o, 'elements', separatedOrTerminatedBy(this.valuePattern(level), atom(','))), v => (o.elements.every(p => p.type === 'PConstant') ? fail : succeed(v))));
        });
    }
    pDict(level) {
        return scope(o => {
            o.type = 'PDict';
            return group('{', mapm(bind(o, 'elements', separatedOrTerminatedBy(value(e => {
                e.value = [];
                return alt(seq(bind(e.value, '0', this.propertyNameExpr()), atom(':'), bind(e.value, '1', this.valuePattern(level))), map(atom(), name => {
                    e.value = [
                        [...DOUBLE_QUOTE, name, ...DOUBLE_QUOTE],
                        { type: 'PConstant', value: [name] },
                    ];
                }));
            }), atom(','))), v => (o.elements.every(e => e[1].type === 'PConstant') ? fail : succeed(v))));
        });
    }
    pConstructor(level, extraStops) {
        return scope(o => {
            o.type = 'PConstructor';
            return seq(bind(o, 'ctor', this.expr(...extraStops, group('(', discard))), 
            // map(anything({ advance: false }), g => o.argumentGroup = g),
            group('(', bind(o, 'arguments', separatedOrTerminatedBy(this.valuePattern(level, ...extraStops), atom(',')))));
        });
    }
    valuePattern(level, ...extraStops) {
        return withoutSpace(alt(scope(o => map(this.pDiscard, _ => o.type = 'PDiscard')), this.pArray(level), this.pDict(level), this.pQuote(level, extraStops), this.pUnquote(level, extraStops), mapm(this.pConstructor(level, extraStops), o => {
            // if (o.arguments.every(a => a.type === 'PConstant')) {
            //     return succeed({
            //         type: 'PConstant',
            //         value: [... o.ctor, o.argumentGroup],
            //     });
            // } else
            if (this.hasCapturesOrDiscards(o.ctor)) {
                const r = match(this.pCaptureDefaultBinder, o.ctor, null, '(');
                if (r !== null && o.arguments.length === 1) {
                    return succeed({
                        type: 'PCapture',
                        inner: o.arguments[0],
                        binder: r
                    });
                }
                else {
                    return fail;
                }
            }
            else {
                return succeed(o);
            }
        }), scope(o => {
            o.type = 'PCapture';
            o.inner = { type: 'PDiscard' };
            return bind(o, 'binder', this.pCaptureBinder(this.binder(...extraStops)));
        }), map(this.expr(...extraStops), e => ({ type: 'PConstant', value: e }))));
    }
}
class SyndicateTypedParser extends SyndicateParser {
    binder(...extraStops) {
        return scope(o => seq(bind(o, 'id', this.identifier), option(seq(atom(':'), bind(o, 'type', this.type(atom('='), ...extraStops))))));
    }
}
//---------------------------------------------------------------------------
// Value pattern utilities
function compilePattern(pattern) {
    const captureBinders = [];
    function walk(pattern) {
        switch (pattern.type) {
            case 'PDiscard':
                return anonymousTemplate `(__SYNDICATE__.QuasiValue._)`;
            case 'PCapture': {
                captureBinders.push(pattern.binder);
                return anonymousTemplate `(__SYNDICATE__.QuasiValue.bind(${walk(pattern.inner)}))`;
            }
            case 'PConstant':
                return anonymousTemplate `(__SYNDICATE__.QuasiValue.litFromJS(${pattern.value}))`;
            case 'PConstructor': {
                const pieces = [anonymousTemplate `${pattern.ctor}`, ...pattern.arguments.map(walk)];
                return anonymousTemplate `(__SYNDICATE__.QuasiValue.ctor(${commaJoin(pieces)}))`;
            }
            case 'PArray': {
                const pieces = pattern.elements.map(walk);
                return anonymousTemplate `(__SYNDICATE__.QuasiValue.arr(${commaJoin(pieces)}))`;
            }
            case 'PDict': {
                const pieces = pattern.elements.map(([k, v]) => anonymousTemplate `[${k}, ${walk(v)}]`);
                return anonymousTemplate `(__SYNDICATE__.QuasiValue.dict(${commaJoin(pieces)}))`;
            }
            case 'PQuote':
                return anonymousTemplate `(__SYNDICATE__.QuasiValue.quote(${walk(pattern.quoted)}))`;
            case 'PUnquote':
                return anonymousTemplate `(__SYNDICATE__.QuasiValue.unquote(${walk(pattern.unquoted)}))`;
            case 'POuterUnquote':
                return anonymousTemplate `(__SYNDICATE__.QuasiValue.unquote(${pattern.outer}))`;
        }
    }
    const skeleton = walk(pattern);
    return {
        skeleton,
        captureBinders,
    };
}

var grammar = /*#__PURE__*/Object.freeze({
    __proto__: null,
    SyndicateParser: SyndicateParser,
    SyndicateTypedParser: SyndicateTypedParser,
    compilePattern: compilePattern
});

/// SPDX-License-Identifier: GPL-3.0-or-later
function stripShebang(items) {
    if ((items.length > 0) &&
        isToken(items[0]) &&
        items[0].text.startsWith('#!')) {
        while (items.length > 0 && !isTokenType(items[0], TokenType.NEWLINE))
            items.shift();
    }
    return items;
}
class ExpansionContext {
    constructor(moduleType, typescript, errorEmitter) {
        this.nextIdNumber = 0;
        this.parser = typescript ? new SyndicateTypedParser : new SyndicateParser();
        this.moduleType = moduleType;
        this.typescript = typescript;
        this.errorEmitter = errorEmitter;
    }
    quasiRandomId() {
        return '__SYNDICATE__id_' + (this.nextIdNumber++);
    }
    argDecl(t, name, type) {
        return (this.typescript) ? t `${name}: ${type}` : t `${name}`;
    }
    emitError(m, loc) {
        this.errorEmitter(m, loc.start, loc.end);
    }
}
function stringifyId(i) {
    return [Object.assign(Object.assign({}, i), { type: TokenType.STRING, text: JSON.stringify(i.text) })];
}
function binderTypeGuard(t) {
    return (binder, index) => {
        if (binder.id.text[0] === '_') {
            return t `${`/* Ignoring underscore-prefixed binder ${binder.id.text} */`}`;
        }
        const raw = t `__vs[${'' + index}]`;
        const bind = t `const ${[binder.id]} = ${raw};`;
        if (binder.type === void 0) {
            return bind;
        }
        else {
            const typeText = itemText(binder.type);
            switch (typeText) {
                case 'boolean':
                case 'string':
                case 'number':
                case 'symbol':
                    return t `if (typeof (${raw}) !== ${JSON.stringify(typeText)}) return;\n${bind}`;
                case 'any':
                    return bind;
                default: {
                    const intermediate = t `__v_${'' + index}`;
                    return t `const ${intermediate} = ${binder.type}.__from_preserve__(${raw});
                             if (${intermediate} === void 0) return;
                             const ${[binder.id]} = ${intermediate};`;
                }
            }
        }
    };
}
function expand(tree, ctx) {
    const macro = new Templates(undefined, { extraDelimiters: ':' });
    function terminalWrap(t, facetToStop, body) {
        if (facetToStop === 'none') {
            return walk(body);
        }
        else {
            const toStop = facetToStop === 'default' ? 'currentSyndicateFacet' :
                facetToStop === 'once-wrapper' ? '__once_facet' :
                    walk(facetToStop);
            const resetCurrentSyndicateFacet = facetToStop === 'once-wrapper' ? [] :
                t `const currentSyndicateFacet = __SYNDICATE__.Turn.activeFacet;`;
            return t `__SYNDICATE__.Turn.active._stop(${toStop}, () => {${resetCurrentSyndicateFacet}${walk(body)}})`;
        }
    }
    function facetWrap(t, facetName, items) {
        if (facetName === 'once-wrapper') {
            return t `__SYNDICATE__.Turn.active.facet(() => {const __once_facet = __SYNDICATE__.Turn.activeFacet; ${items}});`;
        }
        else {
            const defaultLabel = t `const currentSyndicateFacet = __SYNDICATE__.Turn.activeFacet; `;
            const customLabel = facetName === 'default'
                ? []
                : t `const ${facetName.text} = currentSyndicateFacet; `;
            return t `__SYNDICATE__.Turn.active.facet(() => {${defaultLabel}${customLabel}${items}});`;
        }
    }
    function x(p, f) {
        tree = replace(tree, null, p, (v, start) => f(v, macro.template(fixPos(start))));
    }
    function xf(p, f) {
        x(p, (v, t) => t `__SYNDICATE__.Turn.active.${f(v, t)}`);
    }
    const walk = (tree) => expand(tree, ctx);
    const maybeWalk = (tree) => (tree === void 0) ? tree : walk(tree);
    // Unfortunately, because of the incredibly naive repeated
    // traversal of the syntax tree we're doing, the order of the
    // following transformations matters.
    xf(ctx.parser.duringStatement, (s, t) => {
        let spawn = match(ctx.parser.spawn, s.body, null, null);
        if (spawn !== null) {
            if (spawn.linkedToken !== null) {
                ctx.emitError(`during ... spawn doesn't need "linked", it's always linked`, spawn.linkedToken);
            }
            spawn.linkedToken = getRange(s.body);
        }
        let body = (spawn === null)
            ? walk(s.body)
            : expandSpawn(spawn, t, t ` __SYNDICATE__.Turn.activeFacet.preventInertCheck();`);
        const sa = compilePattern(s.pattern);
        const assertion = t `__SYNDICATE__.Observe({
            pattern: __SYNDICATE__.QuasiValue.finish(${sa.skeleton}),
            observer: __SYNDICATE__.Turn.ref(__SYNDICATE__.assertionFacetObserver(
                (${ctx.argDecl(t, '__vs', '__SYNDICATE__.AnyValue')}) => {
                    if (Array.isArray(__vs)) {
${joinItems(sa.captureBinders.map(binderTypeGuard(t)), '\n')}
                        ${body}
                    }
                }
            ))
        })`;
        if (s.test === void 0) {
            return t `assertDataflow(() => ({ target: currentSyndicateTarget(), assertion: ${assertion} }));`;
        }
        else {
            return t `assertDataflow(() => (${walk(s.test)})
                   ? ({ target: currentSyndicateTarget(), assertion: ${assertion} })
                   : ({ target: void 0, assertion: void 0 }));`;
        }
    });
    function expandSpawn(spawn, t, inject = []) {
        // TODO: parentBinders, parentInits
        /*
        let assertions = (s.initialAssertions.length > 0)
            ? t`, new __SYNDICATE__.Set([${commaJoin(s.initialAssertions.map(walk))}])`
            : ``;
        */
        const f = t ` const currentSyndicateFacet = __SYNDICATE__.Turn.activeFacet;`;
        const n = spawn.name === void 0 ? '' : t ` currentSyndicateFacet.actor.name = ${walk(spawn.name)};`;
        return t `__SYNDICATE__.Dataspace._spawn${spawn.linkedToken ? 'Link' : ''}(() => {${f}${n}${inject}${walk(spawn.body)}});`;
    }
    x(ctx.parser.spawn, expandSpawn);
    x(ctx.parser.fieldDeclarationStatement, (s, t) => {
        var _a, _b;
        const ft = ctx.typescript ? t `<${(_a = s.field.type) !== null && _a !== void 0 ? _a : '__SYNDICATE__.AnyValue'}>` : '';
        return t `const ${[s.field.id]} = __SYNDICATE__.Turn.active.field${ft}(${(_b = maybeWalk(s.init)) !== null && _b !== void 0 ? _b : 'void 0'}, ${stringifyId(s.field.id)});`;
    });
    x(ctx.parser.atStatement, (s, t) => {
        return t `(((${ctx.argDecl(t, 'currentSyndicateTarget', '() => __SYNDICATE__.Ref')}) => {${walk(s.body)}})(() => (${walk(s.target)})));`;
    });
    x(ctx.parser.createExpression, (s, t) => {
        return t `__SYNDICATE__.Turn.ref(${walk(s.entity)})`;
    });
    xf(ctx.parser.assertionEndpointStatement, (s, t) => {
        if (s.isDynamic) {
            if (s.test === void 0) {
                return t `assertDataflow(() => ({ target: currentSyndicateTarget(), assertion: ${walk(s.template)} }));`;
            }
            else {
                return t `assertDataflow(() => (${walk(s.test)})
                       ? ({ target: currentSyndicateTarget(), assertion: ${walk(s.template)} })
                       : ({ target: void 0, assertion: void 0 }));`;
            }
        }
        else {
            if (s.test === void 0) {
                return t `assert(currentSyndicateTarget(), ${walk(s.template)});`;
            }
            else {
                return t `replace(currentSyndicateTarget(), void 0, (${walk(s.test)}) ? (${walk(s.template)}) : void 0);`;
            }
        }
    });
    xf(ctx.parser.dataflowStatement, (s, t) => t `_dataflow(() => {${walk(s.body)}});`);
    x(ctx.parser.eventHandlerEndpointStatement, (s, t) => {
        const wrap = s.once ? (i) => facetWrap(t, 'once-wrapper', i) : (i) => i;
        if (s.triggerType === 'dataflow') {
            return wrap(t `__SYNDICATE__.Turn.active._dataflow(() => { if (${walk(s.predicate)}) { ${terminalWrap(t, s.facetToStop, s.body)} } });`);
        }
        if (s.triggerType === 'stop') {
            return t `__SYNDICATE__.Turn.activeFacet.onStop(() => {${walk(s.body)}});`;
        }
        const sa = compilePattern(s.pattern);
        const guardBody = (body) => t `if (Array.isArray(__vs)) {
${joinItems(sa.captureBinders.map(binderTypeGuard(t)), '\n')}
            ${body}
        }`;
        let entity;
        switch (s.triggerType) {
            case 'asserted':
                entity = t `{
                    assert: (${ctx.argDecl(t, '__vs', '__SYNDICATE__.AnyValue')}, ${ctx.argDecl(t, '__handle', '__SYNDICATE__.Handle')}) => {
                        ${guardBody(terminalWrap(t, s.facetToStop, s.body))}
                    }
                }`;
                break;
            case 'retracted':
                entity = t `__SYNDICATE__.assertionObserver((${ctx.argDecl(t, '__vs', '__SYNDICATE__.AnyValue')}) => {
                    ${guardBody(t `return () => { ${terminalWrap(t, s.facetToStop, s.body)} };`)}
                })`;
                break;
            case 'message':
                entity = t `{
                    message: (${ctx.argDecl(t, '__vs', '__SYNDICATE__.AnyValue')}) => {
                        ${guardBody(terminalWrap(t, s.facetToStop, s.body))}
                    }
                }`;
                break;
        }
        const assertion = t `__SYNDICATE__.Observe({
            pattern: __SYNDICATE__.QuasiValue.finish(${sa.skeleton}),
            observer: __SYNDICATE__.Turn.ref(${entity}),
        })`;
        if (s.isDynamic) {
            if (s.test === void 0) {
                return wrap(t `__SYNDICATE__.Turn.active.assertDataflow(() => ({ target: currentSyndicateTarget(), assertion: ${assertion} }));`);
            }
            else {
                return wrap(t `__SYNDICATE__.Turn.active.assertDataflow(() => (${walk(s.test)})
                            ? ({ target: currentSyndicateTarget(), assertion: ${assertion} })
                            : ({ target: void 0, assertion: void 0 }));`);
            }
        }
        else {
            if (s.test === void 0) {
                return wrap(t `__SYNDICATE__.Turn.active.replace(currentSyndicateTarget(), void 0, ${assertion});`);
            }
            else {
                return wrap(t `__SYNDICATE__.Turn.active.replace(currentSyndicateTarget(), void 0, (${walk(s.test)}) ? ${assertion} : void 0);`);
            }
        }
    });
    x(ctx.parser.typeDefinitionStatement, (s, t) => {
        var _a;
        const l = `Symbol.for(${JSON.stringify(s.label.text)})`;
        const fns = JSON.stringify(s.fields.map(f => f.id.text));
        const formatBinder = (b) => { var _a; return t `${[b.id]}: ${(_a = b.type) !== null && _a !== void 0 ? _a : '__SYNDICATE__.AnyValue'}`; };
        const fs = ctx.typescript
            ? t `<{${commaJoin(s.fields.map(formatBinder))}}, __SYNDICATE__.Ref>`
            : '';
        return t `const ${[s.label]} = __SYNDICATE__.Record.makeConstructor${fs}()(${(_a = maybeWalk(s.wireName)) !== null && _a !== void 0 ? _a : l}, ${fns});`;
    });
    xf(ctx.parser.messageSendStatement, (s, t) => t `message(currentSyndicateTarget(), ${walk(s.expr)});`);
    x(ctx.parser.reactStatement, (s, t) => { var _a; return facetWrap(t, (_a = s.label) !== null && _a !== void 0 ? _a : 'default', s.body); });
    x(ctx.parser.stopStatement, (s, t) => t `${terminalWrap(t, s.facetToStop, s.body)};`);
    xf(ctx.parser.syncStatement, (s, t) => t `_sync(${walk(s.peerToSyncWith)}, __SYNDICATE__.Turn.ref({ message() {${walk(s.body)}} }));`);
    return tree;
}
function compile(options) {
    var _a, _b, _c, _d;
    const inputFilename = (_a = options.name) !== null && _a !== void 0 ? _a : '/dev/stdin';
    // console.info(`Syndicate: compiling ${inputFilename}`);
    const source = options.source;
    const moduleType = (_b = options.module) !== null && _b !== void 0 ? _b : 'es6';
    const typescript = (_c = options.typescript) !== null && _c !== void 0 ? _c : false;
    const start = startPos(inputFilename);
    let tree = stripShebang(laxRead(source, { start, extraDelimiters: ':' }));
    // const end = tree.length > 0 ? tree[tree.length - 1].end : start;
    const macro = new Templates(undefined, { extraDelimiters: ':' });
    const ctx = new ExpansionContext(moduleType, typescript, options.emitError);
    tree = expand(tree, ctx);
    const ts = macro.template(fixPos(start));
    // const te = macro.template(fixPos(end));
    {
        const runtime = (_d = options.runtime) !== null && _d !== void 0 ? _d : '@syndicate-lang/core';
        switch (moduleType) {
            case 'es6':
                tree = ts `import * as __SYNDICATE__ from ${JSON.stringify(runtime)};\n${tree}`;
                break;
            case 'require':
                tree = ts `const __SYNDICATE__ = require(${JSON.stringify(runtime)});\n${tree}`;
                break;
            case 'global':
                tree = ts `const __SYNDICATE__ = ${runtime};\n${tree}`;
                break;
            case 'none':
                break;
            default:
                ((_) => {
                    throw new Error(`Unsupported ModuleType: ${moduleType}`);
                })();
        }
    }
    const cw = new CodeWriter(inputFilename);
    cw.emit(tree);
    const text = cw.text;
    return {
        text,
        map: cw.map,
        targetToSourceMap: cw.targetToSourceMap.index(),
        sourceToTargetMap: cw.sourceToTargetMap.index(),
    };
}

var codegen = /*#__PURE__*/Object.freeze({
    __proto__: null,
    stripShebang: stripShebang,
    ExpansionContext: ExpansionContext,
    expand: expand,
    compile: compile
});

export { codegen as Codegen, grammar as Grammar, index as Syntax, compile };
//# sourceMappingURL=syndicate-compiler.es6.js.map
